#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Data   : 2020/08/14 22:59

# 确保安装了前置包
import pyxllib.debug._0_installer
# image部分需要依赖的第三方库
from .installer import *
# pdf相关解析功能
from .fitz_ import *
# 非算法层面的一些简单的图像格式转换功能
from .imlib import *
# 几何运算功能，也有相关的透视变换等图像处理
from pyxllib.cv.cv import *
