#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Data   : 2020/10/18 16:06


"""

跟pytorch有关的一些常用模型、训练的功能封装

"""

from pyxllib.torch.modellib import *
from pyxllib.torch.trainlib import *
